"""determine own version

This only works if build-system.backend-path is respected

:copyright: Copyright (c) 2024 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""

# Actual build system: "_own_version:build_meta"
from setuptools import build_meta as build_meta
import re
import chronver.setuptools_hooks


def __getattr__(name):
    if name == "from_git":
        return chronver.setuptools_hooks.version("chronver")
    raise AttributeError(name)



"""Set up for chronver tests

:copyright: Copyright (c) 2024 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""

import pytest


@pytest.hookimpl(tryfirst=True)
def pytest_runtest_protocol(*args, **kwargs):
    import pathlib, sys

    sys.path.insert(0, str(pathlib.Path(__file__).parent.parent))



"""test git version

:copyright: Copyright (c) 2024 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""

import contextlib


def test_basic():
    from pykern import pkunit
    from chronver import git
    import datetime
    import os
    import time

    with _setup():
        e = datetime.datetime.utcnow().strftime("%Y%m%d.%H%M%S")
        pkunit.pkeq(
            0,
            os.system("git init && echo x > x && git add x && git commit -m x"),
        )
        a = git.version()
        pkunit.pkeq(len(e), len(a))
        pkunit.pkok(e <= a, "expect={} is greater than actual={}", e, a)
        time.sleep(1)
        with open("x", "wt") as f:
            f.write("change")
        b = git.version()
        pkunit.pkok(a < b, "expect={} should be older than={}", a, b)


def test_git_status_error(capsys):
    from pykern import pkunit, pkio, pkdebug
    from chronver import git
    import os
    import subprocess

    with _setup() as d:
        pkio.mkdir_parent(d)
        with pkunit.pkexcept(subprocess.CalledProcessError):
            git.version()
        _, e = capsys.readouterr()
        pkunit.pkre("not a git repository", e)


@contextlib.contextmanager
def _setup():
    from pykern import pkunit, pkio, pkdebug
    from chronver import git
    import os

    with pkunit.save_chdir_work() as d:
        rv = d.join(".git")
        os.environ["GIT_DIR"] = str(rv)
        pkunit.pkeq(None, git.version())
        yield rv



"""test chronver.pkg_info

:copyright: Copyright (c) 2024 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""


def test_basic():
    from pykern import pkunit
    from chronver import pkg_info

    with pkunit.save_chdir_work():
        a = pkg_info.version()
        pkunit.pkeq(None, a)
        e = "20240314.131830"
        with open("PKG-INFO", "wt") as f:
            f.write(
                f"""
Metadata-Version: 2.1
Name: somepkg
Version: {e}
Summary: anything
"""
            )
        pkunit.pkeq(e, pkg_info.version())



"""Get version from PKG-INFO

Used when building a wheel from an sdist, because the git diretory no longer is there.

:copyright: Copyright (c) 2024 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""

import re


def version():
    try:
        with open("PKG-INFO", "rt") as f:
            m = re.search(r"^Version:\s*(\S+)", f.read(), re.MULTILINE)
            if m:
                return m.group(1)
    except FileNotFoundError:
        pass
    return None



"""Chronological version hooks for setuptools builds

:copyright: Copyright (c) 2024 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""

import chronver.git
import chronver.pkg_info


def set_version(dist):
    """Set dist.metadata.version from `version`"""
    dist.metadata.version = version(dist.metadata.name)


def version(name):
    """Try Git and fallback to PKG-INFO

    Returns:
        str: chronological version or raises `LookupError`
    """
    v = chronver.git.version() or chronver.pkg_info.version()
    if v is None:
        raise LookupError(
            f"chronver was unable to detect version for {name}.\n\n"
            "Make sure you're either building from a fully intact git repository "
            "or PyPI tarballs. Most other sources (such as GitHub's tarballs, a "
            "git checkout without the .git folder) don't contain the necessary "
            "metadata and will not work.\n\n"
            "For example, if you're using pip, instead of "
            "https://github.com/user/proj/archive/master.zip "
            "use git+https://github.com/user/proj.git#egg=proj"
        )
    return v



"""Get version from Git

:copyright: Copyright (c) 2024 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""

import datetime
import locale
import os.path
import re
import subprocess
import sys


def version():
    """Chronological version string for most recent commit or current time

    Finds the commit date of git HEAD. Uses ``git status``
    to find files under git control which are modified or to be
    deleted, in which case we assume this is a developer, and we
    should just use the current time for the version. It will be newer
    than any committed version, which is all we care about for
    upgrades.

    Returns:
        str: canonicalized "yyyymmdd.hhmmss" or ``None`` if not a git repo
    """

    def _fmt(value):
        return value.strftime("%Y%m%d.%H%M%S")

    def _head():
        return datetime.datetime.fromtimestamp(
            float(
                _sh(
                    (
                        "git",
                        "log",
                        "-1",
                        "--format=%ct",
                        _sh(("git", "rev-parse", "--abbrev-ref", "HEAD")),
                    ),
                ),
            ),
        )

    def _is_edited():
        """Determine if any files are changed

        Untracked files are hard to know. Could be junk files in
        a developer's repo. This only necessary for development anyway.
        """
        return bool(
            _sh(
                (
                    "git",
                    "status",
                    "--ignore-submodules",
                    "--porcelain",
                    "--untracked-files=no",
                ),
            ),
        )

    def _is_repo():
        return os.path.isdir(".git")

    def _sh(cmd):
        try:
            res = subprocess.check_output(cmd, stderr=subprocess.STDOUT)
            if isinstance(res, bytes):
                return res.decode(locale.getpreferredencoding()).rstrip()
            return None
        except subprocess.CalledProcessError as e:
            if hasattr(e, "output") and len(e.output):
                sys.stderr.write(e.output.decode(locale.getpreferredencoding()))
            raise

    if not _is_repo():
        return None
    return _fmt(datetime.datetime.utcnow() if _is_edited() else _head())



"""test setuptools_hooks

:copyright: Copyright (c) 2024 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""


def test_basic():
    from pykern import pkunit
    from chronver import setuptools_hooks
    import datetime
    import os

    with pkunit.save_chdir_work():
        with pkunit.pkexcept(LookupError):
            setuptools_hooks.version("x")
        e = "20240314.031415"
        with open("PKG-INFO", "wt") as f:
            f.write(
                f"""
Metadata-Version: 2.1
Name: x
Version: {e}
Summary: anything
"""
            )
        pkunit.pkeq(e, setuptools_hooks.version("x"))
        pkunit.pkeq(
            0,
            os.system("git init && echo x > x && git add x && git commit -m x"),
        )
        e = datetime.datetime.utcnow().strftime("%Y%m%d.%H%M%S")
        pkunit.pkok(e, setuptools_hooks.version("x"))
